from __future__ import annotations

from numpy.typing import NDArray

def q_primary(
    production: NDArray, time: NDArray, gain_producer: float, tau_producer: float
) -> NDArray: ...
def q_crm_perpair(injection: NDArray, time: NDArray, gains: NDArray, taus: NDArray) -> NDArray: ...
def q_bhp(pressure_local: NDArray, pressure: NDArray, v_matrix: NDArray) -> NDArray: ...
def calc_A_ij(x_i: float, y_i: float, x_j: float, y_j: float, y_d: float, m: NDArray) -> float: ...
